'''
Fast-Fourier-Transformation

Zeit im Testlauf mit 128x128 Pixel ca. 2 Sekunden.
'''


import matplotlib.pyplot as plt
import matplotlib.colors as colors
import numpy as np
import math
import time


'''
Fast-Fourier-Transformation

Parameter:
    d = -1 fuer Vorwaerts-Transformation (Ortsraums -> Frequenzraum) bzw. d = 1 fuer die Ruecktransformation.
    data ist eine Zeile bzw. Spalte des Bildes (im rekursiven Aufruf entsprechend 1/x der Zeile bzw. Spalte).
    offset ist der Anfangsindex fuer das rekursive Teilstueck.
    k ist die Laenge des rekursiven Teilstuecks.
    s ist die Schrittweite auf dieser Rekursionsstufe.

Anmerkung: Funktioniert in dieser Form nur fuer Bilder deren Groesse eine Zweierpotenz ist.
Siehe auch: http://en.wikipedia.org/wiki/Cooley-Tukey_FFT_algorithm
'''
def fastFourierTransformFunction1DRecursive(d, data, offset, k, s):
    
    # Reihen-Teil fuer die aktuelle Rekursionsstufe anlegen (Laenge k):
    result = np.zeros((k, 2)) # Zweite Dimension von 2 da komplexe Zahlen (0: Realteil, 1: Imaginaerteil)

    # Basisfall k = 1, keine weitere Rekursion noetig. 
    if k == 1:
        result[0] = data[offset]  # Die FT einer Reihe der Laenge 1 ist der Wert selbst.
        
    # Rekursionsfall (k>1), die Reihe muss noch weiter gesplittet werden:
    else:
        # 1. Divide-Schritt:          
        # Rekursive Reihe der halben Laenge (k/2), bestehend aus den Werten mit geradem Index:
        recursiveEven = fastFourierTransformFunction1DRecursive(d, data, offset, k / 2, 2 * s)
        # Rekursive Reihe der halben Laenge (k/2), bestehend aus den Werten mit ungeradem Index:
        recursiveOdd = fastFourierTransformFunction1DRecursive(d, data, offset + s, k / 2, 2 * s)
             
        # 2. Merge-Schritt:       
        # Aus den beiden rekursiv berechneten Reihen der halben Laenge wird nun die Reihe
        # auf dieser Rekursionsebene zusammengesetzt. 
        # Diese hat dann die Laenge k durch Ausnutzung der Periodizitaet/Symmetrie (s.u.):
        for i in xrange(k / 2):
            
            rE = recursiveEven[i] # Nur zwecks Abkuerzung / Lesbarkeit im Folgenden
            rO = recursiveOdd[i] # Nur zwecks Abkuerzung / Lesbarkeit im Folgenden
            
            # Cosinus und Sinus zur Zerlegung von "e hoch ..." der Formel in Realteil und Imaginaerteil:
            # Voberechnen spart ein paar Berechnungen ein, da sin und cos unten jeweils 4 mal auftauchen.
            cosValue = math.cos(d * 2 * math.pi * i / k) # Vorberechnung Cosinus
            sinValue = math.sin(d * 2 * math.pi * i / k) # Vorberechnung Sinus                        
            modRe = (rO[0] * cosValue - rO[1] * sinValue) # Modifikation Realteil
            modIM = (rO[0] * sinValue + rO[1] * cosValue) # Modifikation Imaginaerteil
            
            # Wert der Reihe am Index i berechnen:
            result[i, 0] = rE[0] + modRe
            result[i, 1] = rE[1] + modIM
            
            # Hier wird nun die Periodizitaet/Symmetrie ausgenutzt, d. h. mit den Werten an Index i koennen
            # auch die Werte der Reihe ab der Haelfte bestimmt werden (Index k/2 + i).
            # Man beachte die Subtraktion im Gegensatz zur Addition oben.
            result[k / 2 + i, 0] = rE[0] - modRe
            result[k / 2 + i, 1] = rE[1] - modIM
    
    # Ergebnis zurueckgeben: 
    # Auf oberster Rekursionsebene handelt es sich dabei um eine ganze Zeile bzw. Spalte des Bildes.
    return result
    
    

def fastfourierTransform(data, h, w, d):
        
    result_tmp = np.zeros((h, w, 2))   
    for x in xrange(w):
        print "Working on column", x, "..."
        col = data[:, x]
        result_tmp[:, x] = fastFourierTransformFunction1DRecursive(d, col, 0, h, 1)
       
    result = np.zeros((h, w, 2))     
    for y in xrange(h):
        print "Working on row", y, "..."
        row = result_tmp[y, :]
        result[y, :] = fastFourierTransformFunction1DRecursive(d, row, 0, w, 1)
    
    if d == 1:
        result = result / (w * h)
            
    return result



def fastFourierTransformForward(image):

    h, w = image.shape[0:2]      
    data = np.array([[ (v, 0) for v in row] for row in image])      
    return fastfourierTransform(data, h, w, -1)



def fastFourierTransformBackward(data):

    h, w = data.shape[0:2]       
    return fastfourierTransform(data, h, w, 1)


def shift(data):
    result = np.zeros_like(data)
    result_tmp = np.zeros_like(data)
    h, w = data.shape
    result_tmp[0:h / 2, :] = data[h / 2:h, :]
    result_tmp[h / 2:h, :] = data[0:h / 2, :]
    
    result[:, 0:w / 2] = result_tmp[:, w / 2:w]
    result[:, w / 2:w] = result_tmp[:, 0:w / 2]
    
    return result
    

def aufgabe2():
    input = plt.imread('nerds_128.png')
    
    start = time.time()
    resultForward = fastFourierTransformForward(input)
    end = time.time()
    resultBackward = fastFourierTransformBackward(resultForward)
    
    print "Time taken to compute FFT:", end - start
    
    resultForwardAmp = np.array([[math.sqrt(x*x+y*y) for x, y in row] for row in resultForward])
    resultForwardAmp = shift(np.log(1 + resultForwardAmp))      
    resultBackwardRe = resultBackward[:, :, 0]
 
    plt.figure(1)
    plt.subplot(131)
    plt.title("Original")
    plt.imshow(input, cmap=plt.cm.gray, norm=colors.NoNorm())
    plt.subplot(132)
    plt.title("Forward FFT")
    plt.imshow(resultForwardAmp, cmap=plt.cm.gray)
    plt.subplot(133)
    plt.title("Backward FFT")
    plt.imshow(resultBackwardRe, cmap=plt.cm.gray, norm=colors.NoNorm())
    plt.show()
    


aufgabe2()
