'''
Fast-Fourier-Transformation

Alternative Implementierungsvariante mit Vorberechnung der Basisfunktionen.
Benoetigt damit weniger Berechnungen und ist etwas schneller.
Ansonsten identisch zur Datei "FastFourierFun".

Zeit im Testlauf mit 128x128 Pixel ca. 1,4 Sekunden.

'''


import matplotlib.pyplot as plt
import matplotlib.colors as colors
import numpy as np
import math
import time


'''
Fast-Fourier-Transformation

Parameter:
    table Look-Up-Table fuer: math.cos(d * 2 * PI * i / k) bzw. math.sin(d * 2 * PI * i / k)
    data ist eine Zeile bzw. Spalte des Bildes (im rekursiven Aufruf entsprechend 1/x der Zeile bzw. Spalte).
    offset ist der Anfangsindex fuer das rekursive Teilstueck.
    k ist die Laenge des rekursiven Teilstuecks.
    s ist die Schrittweite auf dieser Rekursionsstufe.
    lvl ist die Rekursionstiefe.

Anmerkung: Funktioniert in dieser Form nur fuer Bilder deren Groesse eine Zweierpotenz ist.
Siehe auch: http://en.wikipedia.org/wiki/Cooley-Tukey_FFT_algorithm
'''
def fastFourierTransformFunction1DRecursive(table, data, offset, k, s, lvl):
    
    # Reihen-Teil fuer die aktuelle Rekursionsstufe anlegen (Laenge k):
    result = np.zeros((k, 2)) # Zweite Dimension von 2 da komplexe Zahlen (0: Realteil, 1: Imaginaerteil)

    # Basisfall k = 1, keine weitere Rekursion noetig. 
    if k == 1:
        result[0] = data[offset]  # Die FT einer Reihe der Laenge 1 ist der Wert selbst.
    # Rekursionsfall (k>1), die Reihe muss noch weiter gesplittet werden:
    else:
        
        kHalf = k / 2
        
        # 1. Divide-Schritt:          
        # Rekursive Reihe der halben Laenge (k/2), bestehend aus den Werten mit geradem Index:
        recursiveEven = fastFourierTransformFunction1DRecursive(table, data, offset, kHalf, 2 * s, lvl + 1)
        # Rekursive Reihe der halben Laenge (k/2), bestehend aus den Werten mit ungeradem Index:
        recursiveOdd = fastFourierTransformFunction1DRecursive(table, data, offset + s, kHalf, 2 * s, lvl + 1)
        
             
        # 2. Merge-Schritt:       
        # Aus den beiden rekursiv berechneten Reihen der halben Laenge wird nun die Reihe
        # auf dieser Rekursionsebene zusammengesetzt. 
        # Diese hat dann die Laenge k durch Ausnutzung der Periodizitaet/Symmetrie (s.u.):
        for i in xrange(kHalf):
                      
            # Cosinus und Sinus zur Zerlegung von "e hoch ..." der Formel in Realteil und Imaginaerteil:
            cosValue = table[lvl, i, 0] # Nachschlagen Cosinus
            sinValue = table[lvl, i, 1] # Nachschlagen Sinus                        
            modRe = (recursiveOdd[i, 0] * cosValue - recursiveOdd[i, 1] * sinValue) # Modifikation Realteil
            modIM = (recursiveOdd[i, 0] * sinValue + recursiveOdd[i, 1] * cosValue) # Modifikation Imaginaerteil
            
            # Wert der Reihe am Index i berechnen:
            result[i, 0] = recursiveEven[i, 0] + modRe
            result[i, 1] = recursiveEven[i, 1] + modIM
            
            # Hier wird nun die Periodizitaet/Symmetrie ausgenutzt, d. h. mit den Werten an Index i koennen
            # auch die Werte der Reihe ab der Haelfte bestimmt werden (Index k/2 + i).
            # Man beachte die Subtraktion im Gegensatz zur Addition oben.
            result[kHalf + i, 0] = recursiveEven[i, 0] - modRe
            result[kHalf + i, 1] = recursiveEven[i, 1] - modIM
    
    # Ergebnis zurueckgeben: 
    # Auf oberster Rekursionsebene handelt es sich dabei um eine ganze Zeile bzw. Spalte des Bildes.
    return result


'''
Erzeugt eine Tabelle mit den Werten der Basisfunktionen, 
in Abhaengigkeit der Parameter, die im rekursiven Algorithmus verwendet werden.
'''
def createLookUpTable(d, n):
    
    t = d * 2 * math.pi # Faktor vorberechnen.
    w = n / 2 # Es gibt maximal halbe Bildbreite bzw. -hoehe verschiedene Werte pro Rekursionsstufe...
    h = int(np.log2(n)) # ...und log2(Breite bzw. Hoehe) viele Rekursionsstufen.
    result = np.zeros((h, w, 2)) # Dritte Dimension ("2") um Realteil und Imaginaerteil gertrennt zu speichern.
    k2 = n / 2
    for k in xrange(0, h):
        # Schleife bis k2, denn auf den unteren Rekursionsstufen braucht es entsprechend weniger Werte.
        # Die Tabelle ist also nicht vollstaendig gefuellt, sondern es werden nur die wirklich
        # benoetigten Eintraege berechnet.
        for i in xrange(k2): 
            cosValue = math.cos(t * i / (2 * k2)) # Realteil
            sinValue = math.sin(t * i / (2 * k2)) # Imaginaerteil
            result[k, i] = (cosValue, sinValue) # In Tabelle ablegen
        k2 = k2 / 2 # Naechste Zeile => Werte fuer naechste Rekursionsstufe => halb so viele Werte
    return result
    

def fastfourierTransform(data, h, w, d):
    
    n = max(h, w)   
    table = createLookUpTable(d, n)
        
    result_tmp = np.zeros((h, w, 2))   
    for x in xrange(w):
        print "Working on column", x, "..."
        col = data[:, x]
        result_tmp[:, x] = fastFourierTransformFunction1DRecursive(table, col, 0, h, 1, 0)
       
    result = np.zeros((h, w, 2))     
    for y in xrange(h):
        print "Working on row", y, "..."
        row = result_tmp[y, :]
        result[y, :] = fastFourierTransformFunction1DRecursive(table, row, 0, w, 1, 0)
    
    if d == 1:
        result = result / (w * h)
            
    return result



def fastFourierTransformForward(image):

    h, w = image.shape[0:2]      
    data = np.array([[ (v, 0) for v in row] for row in image])      
    return fastfourierTransform(data, h, w, -1)



def fastFourierTransformBackward(data):

    h, w = data.shape[0:2]       
    return fastfourierTransform(data, h, w, 1)


def shift(data):
    result = np.zeros_like(data)
    result_tmp = np.zeros_like(data)
    h, w = data.shape
    result_tmp[0:h / 2, :] = data[h / 2:h, :]
    result_tmp[h / 2:h, :] = data[0:h / 2, :]
    
    result[:, 0:w / 2] = result_tmp[:, w / 2:w]
    result[:, w / 2:w] = result_tmp[:, 0:w / 2]
    
    return result
    

def aufgabe2():
    input = plt.imread('nerds_128.png')

    start = time.time()
    resultForward = fastFourierTransformForward(input)
    end = time.time()
    resultBackward = fastFourierTransformBackward(resultForward)
    
    print "Time taken to compute FFT:", end - start
    
    resultForwardAmp = np.array([[math.sqrt(x * x + y * y) for x, y in row] for row in resultForward])
    resultForwardAmp = shift(np.log(1 + resultForwardAmp))      
    resultBackwardRe = resultBackward[:, :, 0]
    
 
    plt.figure(1)
    plt.subplot(131)
    plt.title("Original")
    plt.imshow(input, cmap=plt.cm.gray, norm=colors.NoNorm())
    plt.subplot(132)
    plt.title("Forward FFT")
    plt.imshow(resultForwardAmp, cmap=plt.cm.gray)
    plt.subplot(133)
    plt.title("Backward FFT")
    plt.imshow(resultBackwardRe, cmap=plt.cm.gray, norm=colors.NoNorm())
    plt.show()
    



aufgabe2()
